/*
 * SPDX-FileCopyrightText: 2020-2025 The LineageOS Project
 * SPDX-License-Identifier: Apache-2.0
 */

package org.lineageos.jelly.suggestions

import android.util.Log
import org.json.JSONArray
import org.lineageos.jelly.ext.getCharset
import org.lineageos.jelly.repository.HistoryRepository
import org.json.JSONObject
import java.io.IOException
import java.io.UnsupportedEncodingException
import java.net.HttpURLConnection
import java.net.URL
import java.net.URLEncoder
import java.util.Locale
import java.util.concurrent.TimeUnit

/**
 * The base search suggestions API. Provides common
 * fetching and caching functionality for each potential
 * suggestions provider.
 */
sealed class SuggestionProvider(private val encoding: String) {
    data object Baidu : SuggestionProvider("UTF-8") {
        override fun createQueryUrl(query: String, language: String) =
            "https://suggestion.baidu.com/su?ie=UTF-8&wd=$query&action=opensearch"
    }

    data object Bing : SuggestionProvider("UTF-8") {
        override fun createQueryUrl(query: String, language: String) =
            "https://api.bing.com/osjson.aspx?query=$query&language=$language"
    }

    data object Brave : SuggestionProvider("UTF-8") {
        override fun createQueryUrl(query: String, language: String) =
            "https://search.brave.com/api/suggest?q=$query"
    }

    data object Duck : SuggestionProvider("UTF-8") {
        override fun createQueryUrl(query: String, language: String) =
            "https://duckduckgo.com/ac/?q=$query"

        override fun parseResults(content: String, callback: ResultCallback) {
            val jsonArray = JSONArray(content)
            for (n in 0 until jsonArray.length()) {
                val obj = jsonArray.getJSONObject(n)
                val suggestion = obj.getString("phrase")
                val item = SuggestItem(suggestion)
                if (!callback.addResult(item)) {
                    break
                }
            }
        }
    }

    data object Ecosia : SuggestionProvider("UTF-8") {
        override fun createQueryUrl(query: String, language: String) =
            "https://ac.ecosia.org/autocomplete?q=$query"

        override fun parseResults(content: String, callback: ResultCallback) {
            val mainObject = JSONObject(content)
            val jsonArray = mainObject.getJSONArray("suggestions")
            for (n in 0 until jsonArray.length()) {
                val suggestion = jsonArray.getString(n)
                val item = SuggestItem(suggestion)
                if (!callback.addResult(item)) {
                    break;
                }
            }
        }
    }

    data object Google : SuggestionProvider("UTF-8") {
        override fun createQueryUrl(query: String, language: String) =
            "https://www.google.com/complete/search?client=android&oe=utf8&ie=utf8&cp=4&xssi=t&gs_pcrt=undefined&hl=$language&q=$query"
    }

    data object Qwant : SuggestionProvider("UTF-8") {
        override fun createQueryUrl(query: String, language: String) =
            "https://api.qwant.com/api/suggest?q=$query&lang=$language"
        
        override fun parseResults(content: String, callback: ResultCallback) {
            val mainObject = JSONObject(content)
            val jsonArray = mainObject.getJSONObject("data").getJSONArray("items")
            for (n in 0 until jsonArray.length()) {
                val obj = jsonArray.getJSONObject(n)
                val suggestion = obj.getString("value")
                val item = SuggestItem(suggestion)
                if (!callback.addResult(item)) {
                    break;
                }
            }
        }
    }

    data object Startpage : SuggestionProvider("UTF-8") {
        override fun createQueryUrl(query: String, language: String) =
            "https://www.startpage.com/osuggestions?q=$query"
    }

    data object Yahoo : SuggestionProvider("UTF-8") {
        override fun createQueryUrl(query: String, language: String) =
            "https://search.yahoo.com/sugg/chrome?output=fxjson&command=$query"
    }

    data class History(private val historyRepository: HistoryRepository) :
        SuggestionProvider("UTF-8") {
        override fun createQueryUrl(query: String, language: String) = ""

        override fun fetchResults(rawQuery: String): List<SuggestItem> {
            val list = historyRepository.search(rawQuery, LIMIT)
            return list.map { SuggestItem(it.title, it.url) }
        }
    }

    data object None : SuggestionProvider("UTF-8") {
        override fun createQueryUrl(query: String, language: String) = ""

        override fun fetchResults(rawQuery: String): List<SuggestItem> = emptyList()
    }

    /**
     * Create a URL for the given query in the given language.
     *
     * @param query    the query that was made.
     * @param language the locale of the user.
     * @return should return a URL that can be fetched using a GET.
     */
    protected abstract fun createQueryUrl(query: String, language: String): String

    /**
     * Parse the results of an input stream into a list of [String].
     *
     * @param content  the raw input to parse.
     * @param callback the callback to invoke for each received suggestion
     * @throws Exception throw an exception if anything goes wrong.
     */
    open fun parseResults(
        content: String,
        callback: ResultCallback
    ) {
        val respArray = JSONArray(content)
        val jsonArray = respArray.getJSONArray(1)
        val size = jsonArray.length()
        for (n in 0 until size) {
            val suggestion = jsonArray.getString(n)
            val item = SuggestItem(suggestion)
            if (!callback.addResult(item)) {
                break
            }
        }
    }

    /**
     * Retrieves the results for a query.
     *
     * @param rawQuery the raw query to retrieve the results for.
     * @return a list of history items for the query.
     */
    open fun fetchResults(rawQuery: String): List<SuggestItem> {
        val filter = mutableListOf<SuggestItem>()
        val query = try {
            URLEncoder.encode(rawQuery, encoding)
        } catch (e: UnsupportedEncodingException) {
            Log.e(TAG, "Unable to encode the URL", e)
            return filter
        }

        // There could be no suggestions for this query, return an empty list.
        val content = downloadSuggestionsForQuery(query, language)
            ?.replaceFirst(")]}'", "")
            ?: return filter
        try {
            parseResults(content) {
                filter.add(it)
                filter.size < LIMIT
            }
        } catch (e: Exception) {
            Log.e(TAG, "Unable to parse results", e)
        }
        return filter
    }

    /**
     * This method downloads the search suggestions for the specific query.
     * NOTE: This is a blocking operation, do not fetchResults on the UI thread.
     *
     * @param query the query to get suggestions for
     * @return the cache file containing the suggestions
     */
    private fun downloadSuggestionsForQuery(
        query: String,
        language: String
    ): String? {
        val url = URL(createQueryUrl(query, language))
        val urlConnection = url.openConnection() as HttpURLConnection
        urlConnection.addRequestProperty(
            "Cache-Control",
            "max-age=$INTERVAL_DAY, max-stale=$INTERVAL_DAY"
        )
        urlConnection.addRequestProperty("Accept-Charset", encoding)
        try {
            val charset = urlConnection.getCharset(encoding)
            urlConnection.inputStream.bufferedReader(charset).use {
                return it.readText()
            }
        } catch (e: IOException) {
            Log.d(TAG, "Problem getting search suggestions", e)
        } finally {
            urlConnection.disconnect()
        }
        return null
    }

    companion object {
        private const val TAG = "SuggestionProvider"
        private const val LIMIT = 5
        private val INTERVAL_DAY = TimeUnit.DAYS.toSeconds(1)
        private const val DEFAULT_LANGUAGE = "en"
        private val language by lazy { Locale.getDefault().language.ifEmpty { DEFAULT_LANGUAGE } }

        fun interface ResultCallback {
            fun addResult(suggestion: SuggestItem): Boolean
        }
    }
}
